/*
    siu-sh7343.c - Alsa driver for Renesas' SH7343 SIU peripheral.

    Copyright (c) 2006 Carlos Munoz <carlos@kenati.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <linux/module.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>

#include "siu_sh7343.h"


/* SIU driver dynamic information */
typedef struct {
	u_int32_t	stat;		/* SIU driver object status */
	u_int8_t	open_mode;	/* SIU driver open-mode */
	u_int8_t	rw_flg;		/* stream-data transfer status */
	int32_t		ich;		/* input DMA channel */
	int32_t		och;		/* output DMA channel */
} T_SIU;

typedef struct{
	u_int32_t	**buff;
	u_int32_t	startcnt;	/* transfer start count */
	u_int32_t	undruncnt;	/* buf under-run xfer restart cnt */
	u_int32_t	size;		/* transfer size of stream-data */
	u_int32_t	buf_mtcnt;	/* next source & next destination 
					   transfer stream-data */
	u_int32_t	stactr;		/* transfer start counter */
	u_int32_t	undctr;		/* buf under-run xfer restart cntr */
	u_int32_t	bfwtn;		/* buffer number of written stream-data 
					   (Application -> SIU driver) */
	u_int8_t	buf_mtx;	/* total of stream-data buffer array */
	int		*dercd;		/* error code stored */
	int 		(*callback)(void); /* callback function pointer */
	u_int32_t	**rw_pklev;	/* L-R channel data peak level stored 
					   start address. */
	u_int8_t	pk_mtx;		/* total of peaklevel buffer array */
	u_int32_t	pkchk;		/* input or output data peakcheck 
					   status */
	u_int16_t	peak;		/* peak level of input or output 
					   stream-data */
	u_int32_t	dsl;		/* detect silence status */
	u_int32_t	slcchk;		/* silence check status */
	u_int32_t	interval;	/* period of detect silence by 
					   software.(:second) */
	u_int16_t	level;		/* level of detect silence by 
					   software. */
	int		(*slcchkcallback)(void); /* callback function of 
					   detect silence by software */
	u_int32_t	slccnt;		/* silence time count */
	u_int32_t	slcctr;		/* silence time counter */
} SIU_STMRW;


static snd_card_t		*card;
static u_int8_t			ms;
static u_int32_t		siu_obj_status;
static u_int32_t		infs_tmp;
static u_int32_t		outfs_tmp;

static T_SIU			siu_obj[MAX_SIU_PORTS];
static SIU_STMRW		siu_stmrd[MAX_SIU_PORTS];
static SIU_STMRW		siu_stmwt[MAX_SIU_PORTS];

static u_int32_t		ydef[YRAM_DEF_SIZE];
static u_int32_t		fircoef[YRAM_FIR_SIZE];
static u_int32_t		iircoef[YRAM_IIR_SIZE];

static u_int32_t		stfifo;
static u_int32_t		trdat;


static int snd_siu_sh7343_spb_wait(void)
{
	u_int32_t	cnt;
	u_int32_t	spbwaitcnt;

	spbwaitcnt = 0x00010000;

	/* SPB start by software */
	OUT32(SBACTIV, 0x00000000 );
	for (cnt = 0; cnt < spbwaitcnt; cnt++) {
		if (IN32(SBCTL) == 0x80000000) {
			break;
		}
	}
	if(cnt == spbwaitcnt)
		return -1;

	return 0;
}


static void snd_siu_sh7343_spb_restart(void)
{
	/* SPB hardware start(FIFOCTL source) */
	OUT32(SBACTIV, 0xc0000000);
}


static int snd_siu_sh7343_start( void )
{
	int	i;
	int	off;

	if(siu_obj_status != ST_STOP) {
		printk(KERN_ERR "snd_siu_sh7343_start() invalid status=%d\n",
		       siu_obj_status);
		return -EINVAL;
	}

	/* SIU operates (power down disabled) */
	OUT32(MSTPCR2, IN32(MSTPCR2) & ~MSTPCR2_MSTP208);

	/* Issue software reset to siu */
	OUT32(SRCTL, 0);

	/* Wait for the reset to take effect */
	schedule();

	/* Load initial value to YRAM. Note off is only initialized once */
	for (i = 0, off = 0; i < YRAM0_SIZE; i++, off++)
		ydef[off] = SIU_SPB_PRO_Y0[i];

	for (i = 0; i < YRAM1_SIZE; i++, off++)
		ydef[off] = SIU_SPB_PRO_Y1[i];

	for (i = 0; i < YRAM2_SIZE; i++, off++)
		ydef[off] = SIU_SPB_PRO_Y2[i];

	for (i = 0; i < YRAM3_SIZE; i++, off++)
		ydef[off] = SIU_SPB_PRO_Y3[i];

	for (i = 0; i < YRAM4_SIZE; i++, off++)
		ydef[off] = SIU_SPB_PRO_Y4[i];

	for (i = 0; i < YRAM_FIR_SIZE; i++)
		fircoef[i] = SIU_SRC_44100_44100[i];

	for (i = 0; i < YRAM_IIR_SIZE; i++)
		iircoef[i] = 0x0;

	stfifo = 0x0;
	trdat = 0x0;

	/* portA, portB, SIU operate */
	OUT32(SRCTL, 0x00000301);

	/* portA=256fs, portB=512fs */
	OUT32(CKCTL, 0x40000000);

	/* portA */
	OUT32(BRGASEL, 0x00000000);
	OUT32(BRRA, 0x00000000);

	/* portB */
	OUT32(BRGBSEL, 0x00000000);
	OUT32(BRRB, 0x00000000);

	OUT32(IFCTL, 0x40660000);

	if(ms) {
		/* for Solution Engine */
		OUT16(PKCR, IN16(PKCR) & ~0x3c00);

		/* SIUA_ILR, SIUA_IBT disable */
		OUT16(PKCR, IN16(PKCR) | 0x2800);

		/* MSELCR[2]=B'1 SIUA_OLR, SIUA_OBT output */
		OUT16(MSELCR, IN16(MSELCR) | 0x0004);

		/* portA32bt/fd SIUmaster AportB32bt/fs SIUmaster */
		OUT32(SFORM, 0x0c0c0000);
	} 
	else {
		/* for Solution Engine. SIUA_ILR, SIUA_IBT enable */
		OUT16(PKCR, IN16(PKCR) & ~0x3c00);

		/* SIUA_ILR, SIUA_IBT select */
		OUT16(PSELC, IN16(PSELC) & ~0x1800);

		/* MSELCR[3:0]=B'0000 SIUA_OLR,_OBT,_ILR,_IBT, SIUB_OLR,_OBT,
		   _ILR,_IBT input */
		OUT16(MSELCR, IN16(MSELCR) & ~0x000f);

		/* portA32bt/fs/SIUslave AportB32bt/fs SIUslave */
		OUT32(SFORM, 0x00000000);
	}

	/* SIU driver object information initialize */
	for (i = 0; i < MAX_SIU_PORTS; i++) {
		siu_obj[i].stat = ST_READY;
		siu_obj[i].open_mode = 0x00;
		siu_obj[i].ich = -1;		/* Input DMA channel free */
		siu_obj[i].och = -1;		/* Output DMA channel free */
		siu_obj[i].rw_flg = 0x00;	/* stream-data transfer flag */
		siu_stmrd[i].pkchk = PKCHKOFF;
		siu_stmwt[i].pkchk = PKCHKOFF;
		siu_stmrd[i].dercd = NULL;
		siu_stmwt[i].dercd = NULL;
	}

	infs_tmp = outfs_tmp = 44100;	/* sampling frequency without SRC */
	siu_obj_status = ST_READY;

	return 0;
}


static int snd_siu_sh7343_sampling_rate_convert(u_int32_t	infs, 
						u_int32_t	outfs, 
						u_int32_t	tap,
						u_int32_t	over, 
						u_int32_t	*coef)
{
	int		ercd;
	u_int32_t	cnt;
	u_int32_t	*ptr;
	u_int32_t	*add;
	u_int32_t	tmp;

	if (infs < 8000 || infs > 48000)
		return -EINVAL;
	if (outfs < 8000 || outfs > 48000)
		return -EINVAL;

	if (!(siu_obj_status & ST_OPEN)) {
		if (!(siu_obj_status & ST_READY))
			return -EPERM;
		if (siu_obj_status & ST_SPBACTIV)	/* SPB working? */
			return -EPERM;
		if (tap < 5 || (tap * 2 + 1) > 1023)
			return -EINVAL;
		if ( over < 1 || over > 51)
			return -EINVAL;
		if (coef == NULL)
			return -EINVAL;
		if ((tap * over + 1) > YRAM_FIR_SIZE)
			return -EINVAL;

		tmp = 0x03045000;
		tmp |= (over << 26);
		tmp |= (tap - 4);
		ydef[16] = tmp;

		tmp = (tap * 2 + 1);
		ydef[17] = tmp;
		ydef[20] = outfs;
		ydef[21] = infs;
		/* = log2(over) */
		ydef[22] = (u_int32_t)(log10((double)(over & 0x0000003f)) / 
				       log10(2));
		ydef[27] = (0x40000000/outfs);
		ptr = coef;
		for(cnt = 0; cnt < (tap * over + 1); cnt++) {
			fircoef[cnt] = (u_int32_t)(*ptr++);
		}
	}
	else {
		if (coef) {
			if (tap < 5 || (tap * 2 + 1) > 1023)
				return -EINVAL;
			if (over < 1 || over > 51)
				return -EINVAL;
			if ((tap * over + 1) > YRAM_FIR_SIZE)
				return -EINVAL;
		}

		ydef[20] = outfs;
		ydef[21] = infs;
		ydef[27] = (0x40000000/outfs);

		if (coef) {
			ydef[16] = 0x03045000 | (over << 26) | (tap - 4);
			ydef[17] = (tap * 2 + 1);
			/* = log2(over) */
			ydef[22] = (u_int32_t)
				(log10((double)(over & 0x0000003f)) / log10(2));
		}

		if (siu_obj_status & ST_SPBACTIV) {
			if (infs == outfs) {
				ydef[0] &= ~0x0000000e;	/* SRC OFF */
				ydef[0] |= 0x00000002;	/* FIR through */
			}
			else {
				ydef[0] &= ~0x0000000e;
				ydef[0] |= 0x00000008;	/* SRC ON */
			}

			OUT32(TRDAT, IN32(TRDAT) & ~0x00000001);
			if ((ercd = snd_siu_sh7343_spb_wait()) < 0)
				return ercd;

			/* YRAM variable change */
			add = (u_int32_t *)siu_y_ram;
			*add = ydef[0];
			add = (u_int32_t *)(siu_y_ram + 0x0050);
			*add++ = ydef[20];
			*add++ = ydef[21];
			add = (u_int32_t *)(siu_y_ram + 0x006c);
			*add = ydef[27];

			if (coef) {
				add = (u_int32_t *)(siu_y_ram + 0x0040);
				*add++ = ydef[16];
				*add = ydef[17];
				add = (u_int32_t *)(siu_y_ram + 0x0058);
				*add = ydef[22];
				/* YRAM FIR coefficient */
				ptr = coef;
				add = (u_int32_t *)(siu_y_ram + 0x0200);
				for (cnt = 0; cnt < (tap * over + 1); cnt++) {
					*add++ = (u_int32_t)(*ptr++);
				}
			}
			snd_siu_sh7343_spb_restart();	/* SPB restart */
			OUT32(TRDAT, IN32(TRDAT) | 0x00000001);
		}
	}

	outfs_tmp = outfs;
	infs_tmp = infs;

	siu_obj_status |= ST_SRC_CALL;

	return 0;
}


static int snd_siu_sh7343_graphice_qualizer(u_int32_t	ena,
					    u_int32_t	band,
					    u_int32_t	*coef)
{
	int		ercd;
	u_int32_t	cnt;
	u_int32_t	*ptr;
	u_int32_t	*add;
	u_int32_t	tmp;


	if (siu_obj_status == ST_STOP)
		return -EPERM;
	if (ena != GEQOFF && ena != GEQON)
		return -EINVAL;

	if (ena == GEQOFF) {
		ydef[0] &= ~0x00000020;		/* Graphic EQualizer OFF */
	}
	else {
		if (band < 1 || band > 5)
			return -EINVAL;
		if (coef == NULL)
			return -EINVAL;

		ydef[0] |= 0x00000020;		/* GEQ ON */
		tmp = SIU_SPB_PRO_Y2[0];
		tmp &= ~0x0000003f;
		ydef[48] = tmp | band;		/* SIU_SPB_PRO_Y2[0]; bands */
		ptr = coef;
		for (cnt = 0; cnt < YRAM_IIR_SIZE; cnt++) {
			iircoef[cnt] = (u_int32_t)(*ptr++);
		}
	}

	if (siu_obj_status & ST_SPBACTIV) {
		OUT32(TRDAT, IN32(TRDAT) & ~0x00000001);
		if ((ercd = snd_siu_sh7343_spb_wait()) < 0)
			return ercd;

		/* YRAM variable change */
		add = (u_int32_t *)siu_y_ram;
		*add = ydef[0];
		add = (u_int32_t *)(siu_y_ram + 0x00c0);
		for (cnt = 0; cnt < YRAM2_SIZE; cnt++) {
			*add++ = ydef[48 + cnt];
		}
		if (ena == GEQON) {
			/* XRAM initialization */
			add = (u_int32_t *)(siu_x_ram + 0x0400);
			for (cnt = 0; cnt < XRAM1_SIZE; cnt++) {
				*add++ = 0x0;
			}
			/* YRAM IIR coefficient change */
			add = (u_int32_t *)(siu_y_ram + 0x0600);
			for (cnt=0; cnt < YRAM_IIR_SIZE; cnt++) {
				*add++ = iircoef[cnt];
			}
		}
		snd_siu_sh7343_spb_restart();		/* SPB restart */
		OUT32(TRDAT, IN32(TRDAT) | 0x00000001);
	}

	siu_obj_status |= ST_IIR_CALL;

	return 0;
}


static irqreturn_t siu_sh7343_isr(int			irq, 
				  void			*dev_id,
				  struct pt_regs	*regs)
{
	return IRQ_HANDLED;
}


static int snd_siu_sh7343_init(void)
{
	/* SIUA_OLR, SIUA_OBT, SIUA_OSLD, SIUA_ISLD, SIUMCKA pin setting */
	OUT16(PSELB, IN16(PSELB) & ~(PSELB_PSB0 | PSELB_PSB1));
	OUT16(PSELC, IN16(PSELC) & ~(PSELC_PSC15 | PSELC_PSC14 | PSELC_PSC13));
	OUT16(PKCR, IN16(PKCR) & ~(PKCR_PK0MD0 | PKCR_PK0MD1 | 
				   PKCR_PK1MD0 | PKCR_PK1MD1 | 
				   PKCR_PK2MD0 | PKCR_PK2MD1 | 
				   PKCR_PK3MD0 | PKCR_PK3MD1 | 
				   PKCR_PK4MD0 | PKCR_PK4MD1));

	OUT16(HIZCRB, IN16(HIZCRB) & ~HIZCRB_HIZB4);

	if (IN32(MSTPCR0) & (MSTPCR0_MSTP022 | MSTPCR0_MSTP021)) {
		printk(KERN_ERR "Either DMAC or INTC are powered down\n");
		/* Enable DMAC and INTC */
		OUT32(MSTPCR0, IN32(MSTPCR0) & ~(MSTPCR0_MSTP022 | 
						 MSTPCR0_MSTP021));
	}

	/* Clock supply to SIU halted */
	OUT32(MSTPCR2, IN32(MSTPCR2) | MSTPCR2_MSTP208);

	/* External input clock, B 54MHz */
	OUT32(SCLKACR, SCLKACR_EXSRC);

	return 0;
}


/* Playback and capture have identical properties.  */
static snd_pcm_hardware_t snd_siu_sh7343_pcm_hw =
{
	.info			= (SNDRV_PCM_INFO_MMAP |
				   SNDRV_PCM_INFO_MMAP_VALID |
				   SNDRV_PCM_INFO_INTERLEAVED |
				   SNDRV_PCM_INFO_BLOCK_TRANSFER |
				   SNDRV_PCM_INFO_HALF_DUPLEX),
	.formats		= SNDRV_PCM_FMTBIT_MU_LAW | SNDRV_PCM_FMTBIT_A_LAW,
	.rates			= SNDRV_PCM_RATE_8000_48000,
	.rate_min		= 8000,
	.rate_max		= 48000,
	.channels_min		= 1,
	.channels_max		= 1,
	.buffer_bytes_max	= (64*1024),
	.period_bytes_min	= 1,
	.period_bytes_max	= (64*1024),
	.periods_min		= 1,
	.periods_max		= 1024,
};


static int snd_siu_sh7343_dev_free(snd_device_t	*dev)
{
	/* Nothing to free */
	return 0;
}


static snd_device_ops_t snd_siu_sh7343_dev_ops = {
	.dev_free		= snd_siu_sh7343_dev_free,
};


static int __init alsa_card_siu_sh7343_init(void)
{
        int		err;
	int		i;

	ms = 1;		/* Default is master */
	siu_obj_status = ST_STOP;
	for (i = 0; i < MAX_SIU_PORTS; i++)
		siu_obj[i].stat = ST_STOP;

	/* Initialize siu pins. Siu will be left in power down mode */
	snd_siu_sh7343_init();

	/* Create card */
	if ((card = snd_card_new(0, "siu-sh7343", THIS_MODULE, 0)) == NULL)
		return -ENOMEM;

	strcpy(card->driver, "siu-sh7343");
	strcpy(card->shortname, "Renesas sh7343 SIU");
	sprintf(card->longname, "%s irq %d", card->shortname, SIU_IRQ);

	/* Attach components to the card */
	if ((err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, siu_obj, 
				  &snd_siu_sh7343_dev_ops)) < 0) {
		printk(KERN_ERR "snd_device_new() failed\n");
		goto fail;
	}




	/* Register our isr */
	if ((err = request_irq(SIU_IRQ, siu_sh7343_isr, SA_INTERRUPT, 
					      "SIU", NULL))) {
		printk(KERN_ERR "request_irq() failed\n");
		goto fail;
	}




	/* Now we can register the card instance */
	/*snd_card_register();*/





        return 0;

fail:
	snd_card_free(card);
	return err;
}

static void __exit alsa_card_siu_sh7343_exit(void)
{
	snd_card_free(card);

	free_irq(SIU_IRQ, NULL);
}


MODULE_AUTHOR("Carlos Munoz <carlos@kenati.com>");
MODULE_DESCRIPTION("Alsa SH7343 SIU driver");
MODULE_LICENSE("GPL");

module_init(alsa_card_siu_sh7343_init)
module_exit(alsa_card_siu_sh7343_exit)
